/* copyleft (c) 2003-2021 forum::für::umläute -- IOhannes m zmölnig @ IEM
 * based on d_array.c from pd:
 * Copyright (c) 1997-1921 Miller Puckette and others.
 * For information on usage and redistribution, and for a DISCLAIMER OF ALL
 * WARRANTIES, see the file, "LICENSE.txt," in this distribution.  */

/* shared functions */

#define IEM16_INTERNAL
#include "iem16_table.h"

t_class *table16_class;

void table16_usedindsp(t_table16*x){
  x->x_usedindsp=1;
}

int table16_getarray16(t_table16*x, int*size,t_iem16_16bit**vec){
  *size=x->x_size;
  *vec =x->x_table;
  return 1;
}
